/*
	PixelBlizzard
	Professional sprite animation package.
	
	1994 Chris K. Thomas.  All Rights Reserved.
*/

#include "spriteblizzard.h"
#include "grafutils.h"

short	numSprites;
Sprite*	startSprite;

GWorldPtr	backG;		//background picture
GWorldPtr	constructG;	//where it all comes together

Rect		backRect;	//the part of the background currently showing
Boolean		backNeedsUp;//background needs redrawing

unsigned long			oldTicks;
extern unsigned long	fps;

RgnHandle pxblzPrv_GetSomeSeriousUpdateRgn();
void pxblzprv_AddSprite(Sprite* s);

#pragma segment PixelBlizzard

// * call to initialize
void pxblz_Init(Rect fitInHere,PicHandle back)
{
	startSprite=NULL;
	backNeedsUp=true;
	numSprites=0;
	backG=NULL;
	constructG=NULL;
	backRect=fitInHere;
	pxblz_SetBackground(back);
	oldTicks=TickCount();
	fps=2;
}

/** Sprite Manipulation **/

// * create a new sprite
Sprite *pxblz_NewSprite(short startFrame,short endFrame)
{
	short		numFrames=endFrame-startFrame,i;
	Sprite 		*out=(Sprite*)NewPtrClear(sizeof(Sprite)+(sizeof(Frame) * numFrames));
	CIconHandle	cicon;
	
	if(out==NULL) return NULL;
	if(startSprite==NULL) startSprite=out;
	
	SetGWorld(constructG,GetGWorldDevice(constructG));
	
	out->itsNextSprite=NULL;
	out->itsPrevSprite=NULL;
	out->itsNumFrames=numFrames;
	out->itsMoveProc=NULL;
	out->itsStartFrame=0;
	out->itsEndFrame=0;
	out->itsOldTicks=TickCount();
	out->itsNumTicks=0;
	out->itsIsVisible=false;
	out->itsScaleFactor=0;
	
	for(i=0;i<numFrames;){
		cicon=GetCIcon(startFrame+i);
		if(cicon==NULL){
			DebugStr("\pnot enough memory or no icon found");
		}else{
			RgnHandle	trgn=NULL;
			OSErr		err=noErr;
			BitMap		iconMask;
			PixMapHandle pixMapH;
			
			out->itsFrame[i].itsGWorld=CIconToGWorld(cicon,8);
			if(out->itsFrame[i].itsGWorld==NULL)
				DebugStr("\pGWorld Allocate failed");
			
			pixMapH = out->itsFrame[i].itsGWorld->portPixMap;
			
			iconMask.rowBytes = (**cicon).iconMask.rowBytes;
			iconMask.bounds = (**cicon).iconMask.bounds;
			iconMask.baseAddr = (Ptr)(**cicon).iconMaskData;
			
			trgn=NewRgn();
			err=BitMapToRegion(trgn,&iconMask);
			
			if(trgn==NULL)
				DebugStr("\pGot null mask region.");
			
			if(err!=noErr){
				Str255 yes;
				
				NumToString(err,yes);
				DebugStr(yes);
			}
			
			out->itsFrame[i].itsMask=trgn;
			out->itsOldRect=out->itsRect=(**cicon).iconPMap.bounds;
			DisposeCIcon(cicon);
			
			if(GetGWorldPixMap(out->itsFrame[i].itsGWorld)==NULL)
				DebugStr("\pGot null PixMapHandle after creating frame.");
			
			if(out->itsFrame[i].itsMask==NULL){
				DebugStr("\pGot null mask region after creating frame.");
			}
		//	OffsetRect(&out->itsRect,-out->itsRect.left,-out->itsRect.top);
		}
		++i;
	}
	
	pxblzprv_AddSprite(out);
	return out;
}

// * add sprite to master sprite list
void pxblzprv_AddSprite(Sprite* out)
{
	Sprite *s=startSprite;
	
	if(s==NULL) return;
	
	for(short k=1;k<=numSprites;k++){
		if(s->itsNextSprite==NULL){
			s->itsNextSprite=out;
			out->itsPrevSprite=s;
			break;
		}
		s=s->itsNextSprite;
	}
	
	++numSprites;
}

// * remove the sprite from memory - DO NOT CALL ON ALIASES!
void pxblz_DisposeSprite(Sprite *s)
{
	short i;
	
	if(s->itsNextSprite)
		s->itsNextSprite->itsPrevSprite=s->itsPrevSprite;
	if(s->itsPrevSprite)
		s->itsPrevSprite->itsNextSprite=s->itsNextSprite;
	
	for(i=0;i<s->itsNumFrames;i++){
		DisposeGWorld(s->itsFrame[i].itsGWorld);
		DisposeRgn(s->itsFrame[i].itsMask);
	}
	
	DisposePtr((Ptr)s);
}

// * Move sprite around the game virtual screen (which is as large as the bgknd pict)
void pxblz_MoveSprite(Sprite* s,short h,short v)
{
	short i;
	
	OffsetRect(&s->itsRect,h,v);
	
	for(i=0;i<s->itsNumFrames;i++){
		OffsetRgn(s->itsFrame[i].itsMask,h,v);
	}
}

// * Set the frame range.
void pxblz_SetSpriteFrames(Sprite* s,short from,short to,short current)
{
	s->itsStartFrame=from;
	s->itsEndFrame=to;
	s->itsCurrentFrame=current;
}

// * Set the origins of all sprites to 0,0.
void pxblz_ResetAllSprites()
{
	Sprite *s=startSprite;
	
	for(short k=1;k<=numSprites;k++){
		if(s->itsNextSprite==NULL){
			break;
		}
		
		short	h=s->itsRect.left,
				v=s->itsRect.top;
		OffsetRect(&s->itsRect,-h,-v);
	
		for(short i=0;i<s->itsNumFrames;i++){
			OffsetRgn(s->itsFrame[i].itsMask,-h,-v);
		}
		s=s->itsNextSprite;
	}
	
}

/** Aliases **/

// * New sprite alias
Sprite *pxblz_NewSpriteAlias(Sprite *in)
{
	Sprite 		*out=(Sprite*)NewPtrClear(sizeof(Sprite)+(sizeof(Frame) * in->itsNumFrames));
	
	if(out==NULL) return NULL;
	
	/*copy over or initialize all relevant info*/
	out->itsNextSprite=NULL;
	out->itsPrevSprite=NULL;
	out->itsNumFrames=in->itsNumFrames;
	out->itsMoveProc=in->itsMoveProc;
	out->itsStartFrame=in->itsStartFrame;
	out->itsEndFrame=in->itsEndFrame;
	out->itsOldTicks=TickCount();
	out->itsNumTicks=in->itsNumTicks;
	out->itsIsVisible=false;
	out->itsScaleFactor=in->itsScaleFactor;
	
	for(short i=0;i<out->itsNumFrames;){
		out->itsFrame[i].itsGWorld=in->itsFrame[i].itsGWorld;
		out->itsFrame[i].itsMask=NewRgn();
		CopyRgn(in->itsFrame[i].itsMask,out->itsFrame[i].itsMask);
		out->itsOldRect=out->itsRect=in->itsRect;
		++i;
	}
	
	pxblzprv_AddSprite(out);
	return out;
}

// * Always use this to dispose of aliases!  NEVER just DisposeSprite().
void pxblz_DisposeSpriteAlias(Sprite *s)
{
	if(s->itsNextSprite)
		s->itsNextSprite->itsPrevSprite=s->itsPrevSprite;
	if(s->itsPrevSprite)
		s->itsPrevSprite->itsNextSprite=s->itsNextSprite;
	
	for(short i=0;i<=s->itsNumFrames;i++){
		DisposeRgn(s->itsFrame[i].itsMask);
	}
		
	DisposePtr((Ptr)s);
}

/** Timing **/

// * Do misc. processing tasks
void pxblz_TimeShare(void)
{
	Sprite	*s=startSprite;

	for(;;){
		if(s==NULL) break;
		
		s->itsOldRect=s->itsRect;
		if(s->itsIsVisible){
			if(s->itsMoveProc){
				s->itsMoveProc(s);
			}
			
			if(s->itsStartFrame != s->itsEndFrame){
				if(TickCount() - s->itsOldTicks >= s->itsNumTicks){
					++s->itsCurrentFrame;
					if(s->itsCurrentFrame > s->itsEndFrame){
						s->itsCurrentFrame = s->itsStartFrame;
					}
					s->itsOldTicks = TickCount();
				}
			}//endif(is animated?)
		}		
		
		s=s->itsNextSprite;
	}
}

/** Drawing **/

// * Draw this sprite into constructG
void pxblz_DrawSprite(Sprite* s)
{
	//tRect=s->itsRect;
	//if(s->itsScaleFactor!=0) InsetRect(&tRect,s->itsScaleFactor,s->itsScaleFactor);
	
	if(s->itsCurrentFrame > s->itsNumFrames)
		DebugStr("\pcurrentframe too large");
	
	PixMapHandle pf=GetGWorldPixMap(s->itsFrame[s->itsCurrentFrame].itsGWorld);
	PixMapHandle pc=GetGWorldPixMap(constructG);
	
	CopyBits((BitMap*)&pf,
		(BitMap*)&pc,
		&((GrafPtr)s->itsFrame[s->itsCurrentFrame].itsGWorld)->portRect,
		&s->itsRect,
		srcCopy,
		s->itsFrame[s->itsCurrentFrame].itsMask);
}

// * Draw everything properly
void pxblz_Draw(WindowRef theWindow,Rect theRect)
{
	RgnHandle		tempRgn;
	Sprite*			s=startSprite;
	PixMapHandle	pc=GetGWorldPixMap(constructG);
	PixMapHandle	pb=GetGWorldPixMap(backG);
	
	SetGWorld(backG,NULL);	
	tempRgn=pxblzPrv_GetSomeSeriousUpdateRgn();
		
	CopyBits((BitMap*)&pb,(BitMap*)&pc,
		&backRect,&backRect,srcCopy,tempRgn);
	
	for(;;){
		if(s==NULL) break;
		
		if(s->itsIsVisible==true) pxblz_DrawSprite(s);
		s=s->itsNextSprite;
	}
		
	if(TickCount()-oldTicks>=fps){
		CopyBits((BitMap*)GetGWorldPixMap(constructG),
			&((GrafPtr)theWindow)->portBits,
				&backRect,&theRect,srcCopy,NULL);
		oldTicks=TickCount();
	}
	
	DisposeRgn(tempRgn);
}

// * Update immediately any portion of the constructG
void pxblz_InvalRect(Rect &theRect)
{
	CopyBits(&((GrafPtr)backG)->portBits,&((GrafPtr)constructG)->portBits,
		&theRect,&theRect,srcCopy,NULL);
}

// * Get the update Rgn
RgnHandle pxblzPrv_GetSomeSeriousUpdateRgn()
{
	
	RgnHandle	tRgn=NewRgn();
	
	if(tRgn==NULL) DebugStr("\pNo memory for UpdateRgn.");
	if(MemError()) DebugStr("\pMemory error.");
	OpenRgn();
	
	if(backNeedsUp){
		FrameRect(&((GrafPtr)backG)->portRect);
		backNeedsUp=false;
	}else{
		Sprite	*s=startSprite;
		
		for(;;){
			if(s==NULL) break;
			
			FrameRect(&s->itsOldRect);
			s=s->itsNextSprite;
		}
	}
	
	CloseRgn(tRgn);
	return tRgn;
}

/** Background Manipulation **/

// * setup new background
void pxblz_SetBackground(PicHandle pic)
{
	if(backG!=NULL) DisposeGWorld(backG);
	if(constructG!=NULL) DisposeGWorld(constructG);
	backG=PictureToGWorld(pic,8);
	constructG=PictureToGWorld(pic,8);
	
	/*HLock((Handle)backG->portPixMap);
	HLock((Handle)constructG->portPixMap);*/
	backNeedsUp=true;
}

// * scroll the background around
void pxblz_ScrollBackground(short h,short v)
{
	OffsetRect(&backRect,h,v);
	backNeedsUp=true;
}