/*
	200 pix
*/


#include "CWindowGame.h"
#include "GrafUtils.h"

typedef pascal void (*realCDM)(const BitMap *srcBits, const BitMap *maskBits, const BitMap *dstBits, const Rect *srcRect, const Rect *maskRect, const Rect *dstRect, short mode, RgnHandle maskRgn);
typedef pascal void (*realCB)(const BitMap *srcBits, const BitMap *dstBits, const Rect *srcRect, const Rect *dstRect, short mode, RgnHandle maskRgn);
typedef pascal void (*rectman)(Rect *r, short dh, short dv);
typedef pascal void (*rect1arg)(Rect *r);
typedef pascal void (*noargs)(void);
typedef pascal RgnHandle (*rgnret)(void);
typedef pascal void (*rgn1arg)(RgnHandle);
typedef pascal void (*keys1arg)(KeyMap);

struct{
	realCDM		cdm;	//CopyDeepMask
	realCB		cb;		//CopyBits
	rect1arg	FrameRect;
	rectman		OffsetRect;
	rectman		InsetRect;
	rgnret		NewRgn;
	noargs		OpenRgn;
	rgn1arg		CloseRgn;
	rgn1arg		DisposeRgn;
	keys1arg	GetKeys;
	noargs		SystemTask;
}directTB;

struct {
	short	numLives;
	short	energy;
}Player;

GWorldPtr	backG;
GWorldPtr	spriteG;
GWorldPtr	constructionG;

extern long	fps;
long		oldTicks;
short		currentShot;
const Rect	sptemp={40,40,240,440};
const Rect	playerRrect={1,1,78,43};
const Rect	playerRmaskrect={80,1,157,43};
const Rect	playerLrect={1,177,78,219};
const Rect	playerLmaskrect={80,177,157,219};
const short	kShotLeft=30;
const short	kShotRight=-30;

void ShotDrawSprite(Sprite &s);
void StdDrawSprite(Sprite &s);
void CreateSpriteRgn(RgnHandle *maskRgn,CIconHandle cIconH);

/*** Window ****/
GameWindow::GameWindow()
:DataWindow("\pand you know that notion just crossed my mind",sptemp,FALSE)
{
	UpKey=0x5b;
	DownKey=0x54;
	LeftKey=0x56;
	RightKey=0x58;
	FireKey=0x52;
	
	//CenterOnScreen
	
	directTB.cb=(realCB)GetToolTrapAddress(0xA8EC);
	directTB.cdm=(realCDM)GetToolTrapAddress(0xAA51);
	directTB.InsetRect=(rectman)GetToolTrapAddress(0xA8A9);
	directTB.OffsetRect=(rectman)GetToolTrapAddress(0xA8A8);
	directTB.FrameRect=(rect1arg)GetToolTrapAddress(0xA8A1);
	directTB.NewRgn=(rgnret)GetToolTrapAddress(0xA8D8);
	directTB.OpenRgn=(noargs)GetToolTrapAddress(0xA8DA);
	directTB.CloseRgn=(rgn1arg)GetToolTrapAddress(0xA8DB);
	directTB.DisposeRgn=(rgn1arg)GetToolTrapAddress(0xA8D9);
	directTB.GetKeys=(keys1arg)GetToolTrapAddress(0xA976);
	directTB.SystemTask=(noargs)GetToolTrapAddress(0xA9B4);
	
	ShowReferencedWindow(theWindow);
	LoadEverything();
	SubEventLoop();
}

GameWindow::~GameWindow()
{
	KillEverything();
}

void GameWindow::LoadEverything()
{
	PicHandle	thou;
	PicHandle	back;
	
	oldTicks=TickCount();
	level=(planeOfConfusion**)GetResource('LEVL',128);
	HLockHi((Handle)level);
	
	backNeedsUp=FALSE;
	over=((WindowPtr)theWindow)->portRect;
	back=(PicHandle)GetNamedResource('PICT',"\pLevel 1");
	thou=(PicHandle)GetNamedResource('PICT',"\pthou");
	
	backG=PictureToGWorld(back,8);
	spriteG=PictureToGWorld(thou,8);
	ReleaseResource((Handle)thou);
	constructionG=PictureToGWorld(back,8);
	
	player=NewSprite(playerRrect,playerRmaskrect);
	MoveSprite(player,+0,+110);
	player.itsIsVisible=true;
	
	Rect n2rect={0,0,30,10};
	shot=NewSprite(n2rect,n2rect);
	shot.itsDrawProc=ShotDrawSprite;
	MoveSprite(player,+0,+50);
	
	if(backG==nil){
		DebugStr("\pYes, We Have No RAM...");
	}
	if(spriteG==nil){
		DebugStr("\pYes, We Have No RAM...(2)");
	}
	if(constructionG==nil){
		DebugStr("\pYes, We Have No RAM...(3)");
	}
	
	//cb( &((GrafPtr)backG)->portBits, &((GrafPtr)theWindow)->portBits, &over, &over, srcCopy, NULL);
	SetPort((GrafPtr)theWindow);
	DrawPicture(back,&(**back).picFrame);
	ReleaseResource((Handle)back);
	DrawSprite(player);
}

void GameWindow::KillEverything()
{
	DisposeGWorld(constructionG);
	DisposeGWorld(backG);
	DisposeGWorld(spriteG);
}

void GameWindow::Draw()
{
	PixMapHandle	z;
	RgnHandle		tempRgn;
	
	/*shot processing*/
	if(!shot.itsIsVisible){
		shot.itsRect.top=player.itsRect.top+40;
		shot.itsRect.bottom=shot.itsRect.top+10;
		shot.itsRect.left=player.itsRect.left+40;
		shot.itsRect.right=shot.itsRect.left+20;
	}else{
		MoveSprite(shot,currentShot,0);
		if(currentShot==kShotRight){
			if(shot.itsRect.right>=((GrafPtr)backG)->portRect.right){
				shot.itsIsVisible=false;
				shot.itsOldRect=shot.itsRect;
			}
		}else{
			if(shot.itsRect.left<=((GrafPtr)backG)->portRect.left){
				shot.itsIsVisible=false;
				shot.itsOldRect=shot.itsRect;
			}
		}
	}
	
	/*other processing*/
	tempRgn=GetSomeSeriousUpdateRgn();
	directTB.cb( &((GrafPtr)backG)->portBits, &((GrafPtr)constructionG)->portBits, &over, &over, srcCopy,tempRgn);
	
	z=PreserveGraf((GWorldPtr)constructionG);
	DrawSprite(player);
	DrawSprite(shot);
	RestoreGraf();
	
	if(TickCount()-oldTicks>=fps){
		oldTicks=TickCount();
	}
	directTB.cb( &((GrafPtr)constructionG)->portBits, &((GrafPtr)theWindow)->portBits, &over, &((WindowPtr)theWindow)->portRect, srcCopy,NULL);
	directTB.DisposeRgn(tempRgn);
}

RgnHandle GameWindow::GetSomeSeriousUpdateRgn()
{
	RgnHandle	tRgn=directTB.NewRgn();
	
	directTB.OpenRgn();
	
	if(backNeedsUp){
		directTB.FrameRect(&((WindowPtr)backG)->portRect);
		backNeedsUp=FALSE;
	}else{
		directTB.InsetRect(&player.itsRect,-3,-3);
		directTB.FrameRect(&player.itsRect);
		directTB.InsetRect(&player.itsRect,3,3);
	}
	if(shot.itsIsVisible){
		directTB.FrameRect(&shot.itsOldRect);
	}
	
	directTB.CloseRgn(tRgn);
	return tRgn;
}

void GameWindow::MoveSprite(Sprite& theSprite,short a,short b)
{
	directTB.OffsetRect(&theSprite.itsRect,a,b);
}

Sprite &GameWindow::NewSprite(Rect &gRect,short id)
{
	Sprite out;
	
	out.itsScaleFactor=0;
	out.itsRect=out.itsOldRect=out.itsShapeRect=gRect;
	out.itsDrawProc=StdDrawSprite;
	out.itsIsVisible=false;
	
	CreateSpriteRgn(RgnHandle *maskRgn,CIconHandle cIconH);
	PrimeSprite(out);
	return out;
}

void GameWindow::PrimeSprite(Sprite &s)
{
	PixMapHandle tpix=NewPixMap();
	
	HLockHi(tpix);
	
	CopyBits(&((GrafPtr)spriteG)->portBits,&tpix,tpix);
	DisposePixMap(tpix);
}

// * DrawSprite
//   Draw the sprite and do it's housekeeping.

void GameWindow::DrawSprite(Sprite &s)
{
	//PixMapHandle	implant=PreserveGraf(spriteG);
	
	if(s.itsIsVisible)	s.itsDrawProc(s);
	//RestoreGraf();
}

void StdDrawSprite(Sprite &s)
{
	Rect			tRect;
	
	tRect=s.itsRect;
	if(s.itsScaleFactor!=0) InsetRect(&tRect,s.itsScaleFactor,s.itsScaleFactor);
	directTB.cdm(&((GrafPtr)spriteG)->portBits,&((GrafPtr)spriteG)->portBits,
		&((GrafPtr)constructionG)->portBits,&s.itsShapeRect,&s.itsMaskRect,&tRect,srcCopy,NULL);
	
	s.itsOldRect=tRect;
}

void ShotDrawSprite(Sprite &s)
{
	RGBColor		redc={0xFFFF,0x0000,0x0000};
	
	RGBForeColor(&redc);
	PaintRect(&s.itsRect);
	ForeColor(blackColor);
	
	s.itsOldRect=s.itsRect;
}

void GameWindow::KillSprite(Sprite &s)
{
	//currently, nothing to kill- this will change when we generalize this
	//engine
}


PicHandle GameWindow::GetGWorldPictBounds(GWorldPtr gw,Rect &gRect)
/*
	returns the contents of the GWorld as a picture
*/
{
	PicHandle		result;
	GrafPtr		sp;
	
	GetPort(&sp);
	
	SetPort((GrafPtr)gw);	

	result = OpenPicture(&gw->portRect);
	directTB.cb( &((GrafPtr)gw)->portBits, &((GrafPtr)gw)->portBits, &gRect, &gRect, srcCopy, nil);
	ClosePicture();
	
	SetPort(sp);
	
	return result;
}

void GameWindow::SubEventLoop()
{
	while(!Button()){
		directTB.SystemTask();
		CheckKeys();
		Draw();
	}
}

void GameWindow::CheckKeys()
{
	if(KeyIsDown(UpKey)){
		up();
	}
	if(KeyIsDown(DownKey)){
		down();
	}
	if(KeyIsDown(RightKey)){
		right();
	}
	if(KeyIsDown(LeftKey)){
		left();
	}
	if(KeyIsDown(FireKey)){
		fire();
	}
}

void GameWindow::up()
{
	short	num=(GetHandleSize((Handle)level)/sizeof(planeOfConfusion));
	short	wh=player.itsRect.bottom;
	short	cuml=0;
	
	for(short j=200;j>0;){
		cuml+=1;
		j-=10;
		
		if(wh>=j){
			break;
		}
	}
	
	for(short i=0;i<=num;){
		if(player.itsRect.left<=(**level).gr[i].to && player.itsRect.left>=(**level).gr[i].from){
			switch((**level).gr[i].pixGroup[cuml]){
				case 0:
					MoveSprite(player,-1,-3);
					break;
				case 1:
					break;
				case 2:
					break;
			}
		}
		++i;
	}
}

void GameWindow::down()
{
	MoveSprite(player,+1,+3);
}

void GameWindow::left()
{
	player.itsShapeRect=playerLrect;
	player.itsMaskRect=playerLmaskrect;
	MoveSprite(player,-3,0);
	if(over.right-player.itsRect.right>=300){
		if(over.left!=((GrafPtr)backG)->portRect.left){
			directTB.OffsetRect(&over,-3,0);
			backNeedsUp=TRUE;
		}
	}
}

void GameWindow::right()
{
	player.itsShapeRect=playerRrect;
	player.itsMaskRect=playerRmaskrect;
	MoveSprite(player,+3,0);
	if(over.right-player.itsRect.right<=60){
		if(over.right!=((GrafPtr)backG)->portRect.right){
			directTB.OffsetRect(&over,+3,0);
			backNeedsUp=TRUE;
		}
	}
}

void GameWindow::fire()
{
	/*add code for a shot sprite here*/
	//MoveSprite(player,-3,0);
	shot.itsIsVisible=0x21;
	if(playerRrect.left==player.itsShapeRect.left){
		currentShot= kShotLeft;	//30
	}else{
		currentShot= kShotRight; //-30
	}
}

Boolean GameWindow::KeyIsDown(short keyCode)
{
	union {
		KeyMap			asMap;
		unsigned char	asBytes[16];
	} myMap;
	
	directTB.GetKeys(myMap.asMap);
	return ((myMap.asBytes[keyCode >> 3] >> (keyCode & 0x07)) & 1) != 0;
}

void CreateSpriteRgn(RgnHandle *maskRgn,CIconHandle cIconH)
{
	OSErr err = noErr;
	RgnHandle tempMaskRgn;
	char saveState;
	BitMap iconMask;

	*maskRgn = NULL;

	saveState = HGetState((Handle)cIconH);
	HLock((Handle)cIconH);

	iconMask.rowBytes = (**cIconH).iconMask.rowBytes;
	iconMask.bounds = (**cIconH).iconMask.bounds;
	iconMask.baseAddr = (Ptr)(**cIconH).iconMaskData;

	tempMaskRgn = NewRgn();

	if (tempMaskRgn != NULL){
			err = BitMapToRegion(tempMaskRgn, &iconMask);
		if (err == noErr){
			*maskRgn = tempMaskRgn;
		}else{
			DisposeRgn(tempMaskRgn);
			DebugStr("\pWell?  Where's the RAM!");
		}
	}else{
		DebugStr("\pWell?  Where's the RAM?");
	}
	
	HSetState((Handle)cIconH, saveState);
	return err;
}