#include "offscreen utils.h"

GDHandle	pOldDevice;
CGrafPtr	pOldPort;
/*******************************************************************************

	CreateOffWorld

	Given a rectangle, create an offscreen buffer with the same bounds. If we
	are running on a Mac with the GWorld routines, use them. Otherwise, as
	long as we are running on a Classic QuickDraw machine, whip one up by
	hand. In this little sample, we dont support the creation of color
	offscreen buffers on Color QuickDraw machines that dont have the GWorld
	routines.

*******************************************************************************/
GrafPtr CreateOffWorld(Rect globalRect)
{
	OSErr		err;
	long		value;
	long		height;
	long		width;
//	short		depth;
	BitMap		bitMap;
	GrafPtr		oldPort;
	GDHandle	oldDevice;
	GrafPtr		newPortBeach;
	GWorldPtr	braveNewGWorld;

	err = Gestalt(gestaltQuickdrawVersion, &value);

	if (err == noErr) {
		if (value >= gestalt32BitQD12) {

			err = NewGWorld(&braveNewGWorld,	// result
							0,					// pixel size (0 = find deepest)
							&globalRect,		// bounds
							NULL,				// color table (NIL = default)
							NULL,				// gDevice (not used in this call)
							0);					// flags

			GetGWorld((CGrafPtr*) &oldPort, &oldDevice);
			UseOffWorld((GrafPtr) braveNewGWorld);
			EraseRect(&braveNewGWorld->portRect);
			DoneWithOffWorld((GrafPtr) braveNewGWorld);
			SetGWorld((CGrafPtr) oldPort, oldDevice);

			return (GrafPtr) braveNewGWorld;

		} else if (value == gestaltOriginalQD) {

			height = globalRect.bottom - globalRect.top;
			width = globalRect.right - globalRect.left;

			bitMap.rowBytes = (((width - 1) / 32) + 1) * 4;
			bitMap.bounds = globalRect;
			bitMap.baseAddr = (Ptr) NewHandle(height * bitMap.rowBytes);

			newPortBeach = (GrafPtr) NewPtr(sizeof(GrafPort));
			OpenPort(newPortBeach);
			SetPortBits(&bitMap);
			newPortBeach->portRect = globalRect;
			ClipRect(&globalRect);
			CopyRgn(newPortBeach->clipRgn, newPortBeach->visRgn);

			GetPort(&oldPort);
			SetPort(newPortBeach);
			EraseRect(&braveNewGWorld->portRect);
			SetPort(oldPort);

			return newPortBeach;

		} else {

			DebugStr("\pSorry, unsupported QuickDraw version.");

		}
	}
	return NULL;
}


/*******************************************************************************

	UseOffWorld

	This is sort of like a SetPort routine for our offscreen buffers. If we
	have created a color offscreen buffer with the GWorld routines, lock down
	our pixels and call SetGWorld to set the grafPort and the GDevice. If we
	are running on a B&W machine, lock down our bits and simply call SetPort.
	In both cases, remember the current GrafPort and (if appropriate) GDevice
	so that we can restore them later in DoneWithGWorld.

*******************************************************************************/
void UseOffWorld(GrafPtr offWorlder)
{
	if (IsColorPort(offWorlder)) {
		GetGWorld(&pOldPort, &pOldDevice);
		(void) LockPixels(GetGWorldPixMap((GWorldPtr) offWorlder));
		SetGWorld((CGrafPtr) offWorlder, NULL);
	} else {
		GetPort((GrafPtr*) &pOldPort);
		HLockHi((Handle) offWorlder->portBits.baseAddr);
		offWorlder->portBits.baseAddr = *(Handle) (offWorlder->portBits.baseAddr);
		SetPort(offWorlder);
	}
}


/*******************************************************************************

	DoneWithOffWorld

	This undoes the effects of UseOffWorld. If we are running with the GWorld
	routines, unlock our pixels and reset the GrafPort and GDevice to what
	they were before we called UseOffWorld. If we are running on a B&W
	machine, unlock our bits and restore the old GrafPort.

*******************************************************************************/
void DoneWithOffWorld(GrafPtr offWorlder)
{
	if (IsColorPort(offWorlder)) {
		UnlockPixels(GetGWorldPixMap((GWorldPtr) offWorlder));
		SetGWorld(pOldPort, pOldDevice);
	} else {
		offWorlder->portBits.baseAddr = (Ptr) RecoverHandle(offWorlder->portBits.baseAddr);
		HUnlock((Handle) offWorlder->portBits.baseAddr);
		SetPort((GrafPtr) pOldPort);
	}
}


/*******************************************************************************

	IsColorPort

	Simple utility for determining if the offscreen buffer represents color or
	B&W pixels. Since we only create a color buffer if the GWorld routines are
	available to us, we often use this utility to see if we can call the
	GWorld functions in subsequent routines.

*******************************************************************************/
Boolean IsColorPort(GrafPtr offWorlder)
{
	return (offWorlder->portBits.rowBytes & 0x08000) != 0;
}


/*******************************************************************************

	DisposeOffWorld

	Release the memory for our offscreen buffer. If we created the buffer with
	NewGWorld, call DisposeGWorld. If we created the buffer by hand, we
	dispose of the two blocks of data we created -- the buffer for the
	GrafPort, and the buffer for the bits themselves.

*******************************************************************************/
void DisposeOffWorld(GrafPtr offWorlder)
{
	if (IsColorPort(offWorlder)) {
		DisposeGWorld((GWorldPtr) offWorlder);
	} else {
		DisposeHandle((Handle) offWorlder->portBits.baseAddr);
		ClosePort(offWorlder);
		DisposePtr((Ptr) offWorlder);
	}
}