/*
	PixelBlizzard
	Professional sprite animation package.
	
	1994 Chris K. Thomas.  All Rights Reserved.
*/

#include "spriteblizzard.h"
#include "grafutils.h"
#include "directtoolbox.h"
//#include "HandleStream.h"
#include "iso646.h"

#define debug 0	//set to 1 to enable slowness and debugging checks

short	numSprites;
Sprite*	startSprite;

GWorldPtr	backG;		//background picture
GWorldPtr	constructG;	//where it all comes together

Rect		backRect;	//the part of the background currently showing
Boolean		backNeedsUp;//background needs redrawing
RgnHandle	addWindowRegion;//this is added to the update region when transferring to window

//unsigned long			oldTicks;
//extern unsigned long	fps;

RgnHandle pxblzPrv_GetSomeSeriousUpdateRgn();
void pxblzprv_AddSprite(Sprite* s);

#if (defined __MWERKS__)||(defined applec)
#pragma segment PixelBlizzard
#endif //__MWERKS__

// * call to initialize
void pxblz_Init(Rect fitInHere,PicHandle back)
{
	InitDirectToolbox();
	addWindowRegion=NewRgn();
	startSprite=NULL;
	backNeedsUp=true;
	numSprites=0;
	backG=NULL;
	constructG=NULL;
	backRect=fitInHere;
	pxblz_SetBackground(back);
}

/** Sprite Manipulation **/

// * create a new sprite
Sprite *pxblz_NewSprite(short startFrame,short endFrame)
{
	short		numFrames=endFrame-startFrame,i;
	Sprite 		*out=(Sprite*)NewPtrClear(sizeof(Sprite)+(sizeof(Frame) * numFrames));
	CIconHandle	cicon;
	
	if(out==NULL) return NULL;
	if(startSprite==NULL) startSprite=out;
	
	SetPort((WindowPtr)constructG);
	
	out->itsNextSprite=NULL;
	out->itsPrevSprite=NULL;
	out->itsNumFrames=numFrames;
	out->itsMoveProc=NULL;
	out->itsDrawProc=pxblz_stdDraw;
	out->itsStartFrame=0;
	out->itsEndFrame=0;
	out->itsOldDrawTicks=TickCount();
	out->itsOldMoveTicks=TickCount();
	out->itsDrawTicks=0;
	out->itsMoveTicks=0;
	out->itsIsVisible=false;
	out->itsScaleFactor=0;
	
	for(i=0;i<=numFrames;){
		cicon=GetCIcon(startFrame+i);
		if(cicon==NULL){
			DebugStr("\pnot enough memory or no icon found");
		}else{
			RgnHandle	trgn=NULL;
			OSErr		err=noErr;
			BitMap		iconMask;
			PixMapHandle pixMapH;
			
			out->itsFrame[i].itsGWorld=CIconToGWorld(cicon,0);
			
//###add debug ifdefs around these declarations!
			if(out->itsFrame[i].itsGWorld==NULL)
				DebugStr("\pGWorld Allocate failed");
			
			pixMapH = out->itsFrame[i].itsGWorld->portPixMap;
			HLockHi((Handle)pixMapH);
			
			iconMask.rowBytes = (**cicon).iconMask.rowBytes;
			iconMask.bounds = (**cicon).iconMask.bounds;
			iconMask.baseAddr = (Ptr)(**cicon).iconMaskData;
			
			trgn=NewRgn();
			err=BitMapToRegion(trgn,&iconMask);
			
			if(err!=noErr){
				Str255 yes;
				
				NumToString(err,yes);
				DebugStr(yes);
			}
			
			out->itsFrame[i].itsMask=trgn;
			out->itsOldRect=out->itsRect=(**cicon).iconPMap.bounds;
			DisposeCIcon(cicon);
			
			if(GetGWorldPixMap(out->itsFrame[i].itsGWorld)==NULL)
				DebugStr("\pGot null PixMapHandle after creating frame.");
		//	OffsetRect(&out->itsRect,-out->itsRect.left,-out->itsRect.top);
		}
		++i;
	}
	
	pxblzprv_AddSprite(out);
	return out;
}

// * add sprite to master sprite list
void pxblzprv_AddSprite(Sprite* out)
{
	Sprite *s=startSprite;
	
	if(s==NULL) return;
	
	for(short k=1;k<=numSprites;k++){
		if(s->itsNextSprite==NULL){
			s->itsNextSprite=out;
			out->itsPrevSprite=s;
			break;
		}
		s=s->itsNextSprite;
	}
	
	++numSprites;
}

// * remove the sprite from memory - DO NOT CALL ON ALIASES!
void pxblz_DisposeSprite(Sprite *s)
{
	short i;
	
	if(s->itsNextSprite)
		s->itsNextSprite->itsPrevSprite=s->itsPrevSprite;
	if(s->itsPrevSprite)
		s->itsPrevSprite->itsNextSprite=s->itsNextSprite;
	
	for(i=0;i<=s->itsNumFrames;i++){
		DisposeGWorld(s->itsFrame[i].itsGWorld);
		DisposeRgn(s->itsFrame[i].itsMask);
	}
	
	DisposePtr((Ptr)s);
}

// * Move sprite around the game virtual screen (which is as large as the bgknd pict)
void pxblz_MoveSprite(Sprite* s,short h,short v)
{
	short i;
	
	s->itsIsUpdated=false;
	OffsetRect(&s->itsRect,h,v);
	
	for(i=0;i<=s->itsNumFrames;i++){
		OffsetRgn(s->itsFrame[i].itsMask,h,v);
	}
}

// * Set the frame range.
void pxblz_SetSpriteFrames(Sprite *s,short from,short to,short current)
{
	s->itsStartFrame=from;
	s->itsEndFrame=to;
	s->itsCurrentFrame=current;
}

// * Set the delay (in ticks) between each frame advance and between each call of the
// * moveproc
void pxblz_SetSpriteDelays(Sprite *s,short draws,short moves)
{
	s->itsDrawTicks=draws;
	s->itsMoveTicks=moves;
}

// * Set the origins of all sprites to 0,0.
void pxblz_ResetAllSprites()
{
	Sprite *s=startSprite;
	
	for(short k=1;k<=numSprites;k++){
		if(s->itsNextSprite==NULL){
			break;
		}
		
		short	h=s->itsRect.left,
				v=s->itsRect.top;
		OffsetRect(&s->itsRect,-h,-v);
	
		for(short i=0;i<=s->itsNumFrames;i++){
			OffsetRgn(s->itsFrame[i].itsMask,-h,-v);
		}
		s=s->itsNextSprite;
	}
	
}

/** Aliases **/

// * New sprite alias
Sprite *pxblz_NewSpriteAlias(Sprite *in)
{
	Sprite 		*out=(Sprite*)NewPtrClear(sizeof(Sprite)+(sizeof(Frame) * in->itsNumFrames));
	
	if(out==NULL) return NULL;
	
	/*copy over or initialize all relevant info*/
	
	BlockMove(in,out,sizeof(Sprite));
	out->itsNextSprite=NULL;
	out->itsPrevSprite=NULL;
	
	for(short i=0;i<=out->itsNumFrames;){
		out->itsFrame[i].itsGWorld=in->itsFrame[i].itsGWorld;
		out->itsFrame[i].itsMask=NewRgn();
		CopyRgn(in->itsFrame[i].itsMask,out->itsFrame[i].itsMask);
		out->itsOldRect=out->itsRect=in->itsRect;
		++i;
	}
	
	pxblzprv_AddSprite(out);
	return out;
}

// * Always use this to dispose of aliases!  NEVER just DisposeSprite().
void pxblz_DisposeSpriteAlias(Sprite *s)
{
	if(s->itsNextSprite)
		s->itsNextSprite->itsPrevSprite=s->itsPrevSprite;
	if(s->itsPrevSprite)
		s->itsPrevSprite->itsNextSprite=s->itsNextSprite;
	
	for(short i=0;i<=s->itsNumFrames;i++){
		DisposeRgn(s->itsFrame[i].itsMask);
	}
		
	DisposePtr((Ptr)s);
}

/** Timing **/

// * Do misc. processing tasks
void pxblz_TimeShare(void)
{
	Sprite	*s=startSprite;
	Rect	*temp=NULL;
	
	
	for(;;){
		if(s==NULL) break;
		
		s->itsOldRect=s->itsRect;
		if(s->itsIsVisible){
			if(s->itsMoveProc && TickCount() - s->itsOldMoveTicks >= s->itsMoveTicks){
				if(temp and s->itsCollideProc and SectRect(&s->itsRect,temp,temp)){
					s->itsCollideProc(s);
				}
				s->itsMoveProc(s);
				s->itsOldMoveTicks=TickCount();
				temp=&s->itsRect;
			}
			
			if(s->itsStartFrame != s->itsEndFrame){
				if(TickCount() - s->itsOldDrawTicks >= s->itsDrawTicks){
					if(++s->itsCurrentFrame > s->itsEndFrame){
						s->itsCurrentFrame = s->itsStartFrame;
					}
					s->itsOldDrawTicks = TickCount();
				}
			}//endif(is animated?)
		}		
		
		s=s->itsNextSprite;
	}
}

/** Drawing **/

// * Draw this sprite into constructG
void pxblz_DrawSprite(Sprite *s)
{
	//tRect=s->itsRect;
	//if(s->itsScaleFactor!=0) InsetRect(&tRect,s->itsScaleFactor,s->itsScaleFactor);
	
	pxblz_LockGWorld(s->itsFrame[s->itsCurrentFrame].itsGWorld);
	SetPort((GrafPtr)s->itsFrame[s->itsCurrentFrame].itsGWorld);
#if debug	
	if(s->itsCurrentFrame>s->itsNumFrames)
		DebugStr("\pcurrentframe too large");
#endif
	if(s->itsDrawProc){
		s->itsDrawProc(s);
	}
	
	pxblz_UnlockGWorld(s->itsFrame[s->itsCurrentFrame].itsGWorld);
}

void pxblz_stdDraw(Sprite *s)
{
	directTB.CopyBits(&((GrafPtr)s->itsFrame[s->itsCurrentFrame].itsGWorld)->portBits,
			&((GrafPtr)constructG)->portBits,
			&((GrafPtr)s->itsFrame[s->itsCurrentFrame].itsGWorld)->portRect,
			&s->itsRect,
			srcCopy,
			s->itsFrame[s->itsCurrentFrame].itsMask);
}

void pxblz_fxDraw(Sprite *s)
{
	short mode;
	
	switch(s->itsMiscData){
		case 1:
			mode=notSrcCopy;
			break;
		default:
			mode=srcCopy;
			break;
	}
	
	directTB.CopyBits(&((GrafPtr)s->itsFrame[s->itsCurrentFrame].itsGWorld)->portBits,
		&((GrafPtr)constructG)->portBits,
		&((GrafPtr)s->itsFrame[s->itsCurrentFrame].itsGWorld)->portRect,
		&s->itsRect,
		mode,
		s->itsFrame[s->itsCurrentFrame].itsMask);
}

// * Draw everything properly
void pxblz_Draw(WindowRef theWindow,Rect theRect)
{
	Sprite*			s=startSprite;
	Rect			spareRect;
	
	pxblz_LockGWorld(backG);
	pxblz_LockGWorld(constructG);
	
	SetPort((WindowPtr)constructG);
	
	s=startSprite;
	
	for(;;){

		if(s==NULL) break;

		if(s->itsIsVisible && !s->itsIsUpdated){
			
			//copy the background into the construction gworld
			//to erase our previous image
			
			UnionRect(&s->itsRect,&s->itsOldRect,&spareRect);
			directTB.CopyBits(&((GrafPtr)backG)->portBits,
				&((GrafPtr)constructG)->portBits,
					&spareRect,&spareRect,srcCopy,NULL);
			
			//draw the sprite into the construction gworld
			pxblz_DrawSprite(s);
			
			//put the construction gworld up on the screen
			spareRect=s->itsOldRect;
			OffsetRect(&spareRect,-backRect.left,theRect.top);
			directTB.CopyBits(&((GrafPtr)constructG)->portBits,
				&((GrafPtr)theWindow)->portBits,
					&s->itsOldRect,&spareRect,srcCopy,NULL);
			
			spareRect=s->itsRect;
			OffsetRect(&spareRect,-backRect.left,theRect.top);
			
			directTB.CopyBits(&((GrafPtr)constructG)->portBits,
				&((GrafPtr)theWindow)->portBits,
					&s->itsRect,&spareRect,srcCopy,NULL);
			
			s->itsIsUpdated=true;
			s->itsOldRect=s->itsRect;
		}
		s=s->itsNextSprite;
	}
	
	/*Copy the construction GWorld onto the screen*/
	if(backNeedsUp)	{
		directTB.CopyBits(&((GrafPtr)constructG)->portBits,
		&((GrafPtr)theWindow)->portBits,
			&backRect,&theRect,srcCopy,NULL);
		
		backNeedsUp=false;
	}

	
	pxblz_UnlockGWorld(backG);
	pxblz_UnlockGWorld(constructG);
}

// * Update immediately any portion of the constructG
void pxblz_InvalRect(Rect &theRect)
{
	pxblz_LockGWorld(backG);
	pxblz_LockGWorld(constructG);
	
	directTB.CopyBits(&((GrafPtr)backG)->portBits,&((GrafPtr)constructG)->portBits,
		&theRect,&theRect,srcCopy,NULL);
	
	pxblz_UnlockGWorld(backG);
	pxblz_UnlockGWorld(constructG);
}

// * Get the update Rgn
RgnHandle pxblzPrv_GetSomeSeriousUpdateRgn()
{
	SetPort((GrafPtr)constructG);
	
	RgnHandle	tRgn=NewRgn();
	Rect		nuRect;

#if debug
	if(tRgn==NULL) DebugStr("\pAin't No UpdateRgn");
	if(MemError()) DebugStr("\pMemory error.");
#endif

	OpenRgn();
	
	Sprite	*s=startSprite;
	
	for(;;){
		if(s==NULL) break;
		
		if(s->itsIsVisible==true){
/*				RgnHandle oneRgn=s->itsFrame[s->itsCurrentFrame].itsMask;
			InsetRgn(oneRgn,-1,-1);
			FrameRgn(oneRgn);
			InsetRgn(oneRgn,1,1);*/
			UnionRect(&s->itsOldRect,&s->itsRect,&nuRect);
			FrameRect(&nuRect);
		}
		s=s->itsNextSprite;
	}
	
	CloseRgn(tRgn);
	return tRgn;
}

// * Lock down the specified gworld's pixmap
void pxblz_LockGWorld(GWorldPtr theG)
{
	PixMapHandle ourPix=GetGWorldPixMap(theG);
	
#if debug
	if(ourPix==NULL)
		DebugStr("\pLockGWorld: ourPix is NULL");
#endif
	LockPixels(ourPix);
}

// * Unlock the specified gworld's pixmap
void pxblz_UnlockGWorld(GWorldPtr theG)
{
	PixMapHandle ourPix=GetGWorldPixMap(theG);
	
#if debug
	if(ourPix==NULL)
		DebugStr("\pUnlockGWorld: ourPix is NULL");
#endif
	UnlockPixels(ourPix);
}

/** Background Manipulation **/

// * setup new background
void pxblz_SetBackground(PicHandle pic)
{
	if(backG!=NULL) DisposeGWorld(backG);
	if(constructG!=NULL) DisposeGWorld(constructG);
	backG=PictureToGWorld(pic,0);
	constructG=PictureToGWorld(pic,0);
	
	HLock((Handle)backG->portPixMap);
	HLock((Handle)constructG->portPixMap);
	backNeedsUp=true;
}

// * scroll the background around
void pxblz_ScrollBackground(short h,short v,Rect theRect)
{
	OffsetRect(&backRect,h,v);
	backNeedsUp=true;
}