#pragma once
/*
	PixelBlizzard
	1994 Chris K. Thomas.  All Rights Reserved.
*/
#include "WindowExtensions.h"

typedef struct{
	GWorldPtr	itsGWorld;
	RgnHandle	itsMask;
}Frame;

typedef unsigned long ulong;

struct Sprite{
	Sprite*	itsNextSprite;		//next sprite in list
	Sprite*	itsPrevSprite;		//prev sprite in list
	Rect	itsRect;			//rect on screen
	Rect	itsOldRect;			//Update rect
	
	Boolean	itsIsVisible;		//Is visible on screen?
	Boolean	itsIsUpdated;		//Has been drawn on screen since last moved?
	
	short	itsScaleFactor;		//how much (+/-) to scale the picture on screen? (does not hittest!)
									//NOTE: this is unused at present
	ulong	itsOldDrawTicks;	//TickCount() just after last frame advance
	ulong	itsOldMoveTicks;	//TickCount() just after last move
	short	itsMoveTicks;		//Number of ticks to wait between itsMoveProc calls
	short	itsDrawTicks;		//number of ticks before advancing frame
	short	itsNumFrames;		//number of total frames
	short	itsStartFrame;		//starting frame to display
	short	itsEndFrame;		//ending frame
	short	itsCurrentFrame;	//the actual frame currently displayed
	void	(*itsMoveProc)(Sprite*);//move your own rectangle
	void	(*itsDrawProc)(Sprite*);//draw yourself (everything is locked)
	void	(*itsCollideProc)(Sprite*);//what to do when you collide
	long	itsMiscData;		//data- should probably only be used legit. by sprite aliases
	Frame	itsFrame[1];		//the frame data
};

typedef struct Sprite Sprite;

/** inline utilities **/
inline short RectSizeH(Rect it)
{
	return (it.right-it.left);
}

inline short RectSizeV(Rect it)
{
	return (it.bottom-it.top);
}

/** Init **/
void pxblz_Init(Rect fitInHere,PicHandle back);

/** Sprite Manipulation **/
Sprite *pxblz_NewSprite(short startFrame,short endFrame);
void pxblz_DisposeSprite(Sprite* s);
void pxblz_MoveSprite(Sprite* s,short h,short v);
void pxblz_SetSpriteFrames(Sprite* s,short from,short to,short current);
void pxblz_SetSpriteDelays(Sprite *s,short draws,short moves);
void pxblz_ResetAllSprites();

/** Sprite Aliases **/
Sprite *pxblz_NewSpriteAlias(Sprite *in);
void pxblz_DisposeSpriteAlias(Sprite *s);

/** Timing **/
void pxblz_TimeShare(void);

/** Drawing **/
void pxblz_DrawSprite(Sprite* s);
void pxblz_Draw(WindowRef theWindow,Rect theRect);
void pxblz_stdDraw(Sprite *s);
void pxblz_fxDraw(Sprite *s);
void pxblz_InvalRect(Rect &theRect);
RgnHandle pxblzPrv_GetSomeSeriousUpdateRgn();
void pxblz_LockGWorld(GWorldPtr theG);
void pxblz_UnlockGWorld(GWorldPtr theG);

/** Background Manipulation **/
void pxblz_SetBackground(PicHandle pic);
void pxblz_ScrollBackground(short h,short v,Rect theRect);